require 'extant/attribute'
require 'extant/attribute_definition'

module Extant
  module Attributes
    def self.included(base)
      base.extend(ClassMethods)
    end

    module ClassMethods
      def attribute_definitions
        @attributes ||= {}
      end

      def attribute(name, type=nil, **opts)
        opts[:type] = type
        attribute_definitions[name] = AttributeDefinition.new(self, name, opts)
      end
    end

    def initialize(attrs={})
      attrs = Hash[attrs.map{ |k, v| [k.to_sym, v] }]
      attribute_definitions.each do |k, definition|
        value = if attrs.key?(k)
          attrs[k]
        else
          UnsetValue
        end

        if UnsetValue == value && definition.has_default
          value = definition.default_value.is_a?(Proc) ? definition.default_value.call : definition.default_value
        end

        extant_attributes[k] =
          Extant::Attribute.new(k, value, definition)
      end

      self.original_extant_attributes = extant_attributes
    end

    def attributes
      self.class.attribute_definitions.each_with_object({}) do |(k, _), attrs|
        attrs[k] = send(k)
      end
    end

    def attribute_definitions
      self.class.attribute_definitions
    end

    def extant_attributes
      @extant_attributes ||= {}
    end

    def set_attributes
      self.class.attribute_definitions.each_with_object({}) do |(k, _), attrs|
        if extant_attributes[k].set?
          attrs[k] = send(k)
        end
      end
    end

    def changed_attributes
      self.class.attribute_definitions.each_with_object({}) do |(k, _), attrs|
        if extant_attributes[k].has_changed?
          attrs[k] = send(k)
        end
      end
    end

    def each_extant_attribute
      extant_attributes.values.each do |v|
        yield v
      end
    end

    def update_attributes(attrs={})
      attrs.each do |k, v|
        send("#{k}=".to_sym, v) if respond_to?("#{k}=".to_sym)
      end

      self
    end

    def dirty?(attr=nil)
      if attr
        !!changed_extant_attributes[attr]
      else
        changed_extant_attributes.any?
      end
    end

    def original_extant_attributes
      @original_extant_attributes ||= {}
    end

    def with(attrs={}, metadata={})
      new_instance = self.class.new(self.attributes)
      new_instance.update_attributes(attrs)

      new_instance.send(:with_metadata=, metadata)
      new_instance.send(:original_extant_attributes=, original_extant_attributes)
      new_instance
    end

    def with_metadata
      @with_metadata || {}
    end

    private

    attr_writer :with_metadata

    def changed_extant_attributes
      self.class.attribute_definitions.each_with_object({}) do |(k, _), attrs|
        if extant_attributes[k].has_changed?
          attrs[k] = extant_attributes[k]
        end
      end
    end

    def original_extant_attributes=(attrs={})
      attrs.each do |k, v|
        original_extant_attributes[k] = v.dup
      end
    end
  end
end

require 'extant/coercers'
