module Extant
  module Coercers
    def self.find(type)
      if type.is_a?(::Hash)
        coercer_options = {
          hash_format: type
        }

        type = HashCoercerBuilder.build(coercer_options)
      end

      if type.is_a?(::Array)
        coercer_options = {
          array_format: type
        }

        type = ArrayCoercerBuilder.build(coercer_options)
      end

      if type.is_a?(Class) && type.ancestors.include?(Extant::Coercers::Base)
        coercer_class = type
      else
        coercer_class = Object.const_get("Extant::Coercers::#{type}")

        unless coercer_class.method_defined?(:coerce)
          raise "#{type} must respond to #coerce"
        end
      end

      coercer_class
    end
  end
end

require 'extant/coercers/base'

require 'extant/coercers/boolean'
require 'extant/coercers/date_time'
require 'extant/coercers/fixnum'
require 'extant/coercers/float'
require 'extant/coercers/hash'
require 'extant/coercers/array'
require 'extant/coercers/integer'
require 'extant/coercers/string'
require 'extant/coercers/symbol'
require 'extant/coercers/time'
require 'extant/coercers/uuid'
