module Extant::Coercers
  class Array < Base
    class << self
      attr_accessor :value_type
    end

    def coerce
      return ArgumentError unless value.respond_to?(:to_a)
      result = value.to_a
      return ArgumentError unless result.is_a?(::Array)

      if value_type
        result = result.map do |v|
          coerce_value(v)
        end
      end

      self.coerced = true

      result
    rescue ArgumentError
      UncoercedValue
    end

    private

    def coerce_value(v)
      coercer = value_type.new(v)
      result = coercer.coerce
      raise ArgumentError unless coercer.coerced?
      result
    end

    def key_type
      self.class.key_type
    end

    def value_type
      self.class.value_type
    end
  end

  class ArrayCoercerBuilder
    def self.build(opts)
      value_type_local = Extant::Coercers.find(opts[:array_format].first)

      Class.new(Extant::Coercers::Array) do
        self.value_type = value_type_local

        def self.coercer_name
          'array'
        end
      end
    end
  end
end
