module Extant::Coercers
  class Hash < Base
    class << self
      attr_accessor :key_type, :value_type
    end

    def coerce
      result = ::Hash[value]

      if key_type && value_type
        result = result.each_with_object({}) do |(k, v), h|
          h[coerce_key(k)] = coerce_value(v)
        end
      end

      self.coerced = true

      result
    rescue ArgumentError
      UncoercedValue
    end

    private

    def coerce_key(v)
      coercer = key_type.new(v)
      result = coercer.coerce
      raise ArgumentError unless coercer.coerced?
      result
    end

    def coerce_value(v)
      coercer = value_type.new(v)
      result = coercer.coerce
      raise ArgumentError unless coercer.coerced?
      result
    end

    def key_type
      self.class.key_type
    end

    def value_type
      self.class.value_type
    end
  end

  class HashCoercerBuilder
    def self.build(opts)
      key_type_local = Extant::Coercers.find(opts[:hash_format].keys.first)
      value_type_local = Extant::Coercers.find(opts[:hash_format].values.first)

      Class.new(Extant::Coercers::Hash) do
        self.key_type = key_type_local
        self.value_type = value_type_local

        def self.coercer_name
          'hash'
        end
      end
    end
  end
end
